@@ -1,7 +1,9 @@
+from hashlib import sha512
 import cPickle as pickle
 import os
 import sys
 import logging
+import pkg_resources
 import subprocess
 from contextlib import contextmanager
 from urlparse import urlparse
@@ -9,6 +11,7 @@ from . import command
 from . import os_api
 from . import virtualenv_api
 from .environment_utils import EnvironmentUtils
+from .installer import Installer
 from .checkout_cache import CheckoutCache
 from .sources import (
     get_all_sources_dict,
@@ -22,30 +25,58 @@ class PythonEnvironment(object):
     def __init__(self, argv=()):
         super(PythonEnvironment, self).__init__()
         self._argv = list(argv)
+        self.installer = Installer(self)
         self.utils = EnvironmentUtils(self)
+        self._aliases = {}
+        self._executed_deployment_file_locations = set()
+        self._executed_deployment_file_hashes    = set()
+    def add_alias(self, name, source):
+        self._aliases[name] = source
+    def has_alias(self, name):
+        if isinstance(name, pkg_resources.Requirement):
+            name = name.unsafe_name
+        return name in self._aliases
     def checkout(self, source, *args, **kwargs):
+        """
+        Downloads a package from *source*, and returns the path to which it was extracted.
+        """
         source = self._make_source_object(source)
         _logger.info("Checking out %r (%s)", source.get_name(), type(source).__name__)
         source = self._make_source_object(source)
         return source.checkout(self, *args, **kwargs)
     def create_and_activate(self):
         raise NotImplementedError() # pragma: no cover
+    def execute_deployment_file_once(self, path_or_url):
+        executed = False
+        if path_or_url not in self._executed_deployment_file_locations:
+            content = self._get_deployment_file_content(path_or_url)
+            content_hash = sha512(content).digest()
+            if content_hash not in self._executed_deployment_file_hashes:
+                self.execute_deployment_string(content)
+                self._executed_deployment_file_hashes.add(content_hash)
+                executed = True
+            self._executed_deployment_file_locations.add(path_or_url)
+        if not executed:
+            _logger.info("Deployment file at %r already executed. Skipping...", path_or_url)
     def execute_deployment_file(self, path_or_url):
-        if self._is_url(path_or_url):
-            self.execute_deployment_file_url(path_or_url)
-        else:
-            self.execute_deployment_file_path(path_or_url)
+        self.execute_deployment_string(self._get_deployment_file_content(path_or_url))
     def execute_deployment_file_url(self, url):
         _logger.info("Executing deployment file from url %r...", url)
         self.execute_deployment_file_object(os_api.urlopen(url))
     def execute_deployment_file_object(self, fileobj, filename=None):
+        return self.execute_deployment_string(fileobj.read(), filename=filename)
+    def execute_deployment_string(self, content, filename=None):
         executed_locals = self._get_config_file_locals(filename)
-        exec fileobj.read() in executed_locals
+        exec content in executed_locals
     def execute_deployment_file_path(self, path):
         _logger.info("Executing deployment file %r...", path)
         with open(path, "rb") as fileobj:
             with self._filename_as_argv0(path):
                 self.execute_deployment_file_object(fileobj, filename=path)
+    def _get_deployment_file_content(self, path_or_url):
+        if self._is_url(path_or_url):
+            return os_api.urlopen(path_or_url).read()
+        return open(path_or_url).read()
     def execute_deployment_stdin(self):
         _logger.info("Executing deployment script from standard input...")
         self.execute_deployment_file_object(sys.stdin, filename="<stdin>")
@@ -59,12 +90,18 @@ class PythonEnvironment(object):
             self._argv = old_argv
 
     def install(self, source, reinstall=True):
-        source = self._make_source_object(source)
-        _logger.info("Installing %r (%s)", source.get_name(), type(source).__name__)
-        if not self._is_already_installed(source) or reinstall:
-            returned = source.install(self)
-            self._post_install(source)
-            return returned
+        """
+        Installs a package from *source*.
+
+        @param reinstall: If False, skips installation if this source has already been installed.
+        """
+        with self.installer.get_installation_context():
+            source = self._make_source_object(source)
+            _logger.info("Installing %r (%s)", source.get_name(), type(source).__name__)
+            if not self._is_already_installed(source) or reinstall:
+                returned = source.install(self, reinstall=reinstall)
+                self._post_install(source)
+                return returned
     def get_argv(self):
         return list(self._argv)
     def get_checkout_cache(self):
@@ -73,9 +110,9 @@ class PythonEnvironment(object):
         raise NotImplementedError() # pragma: no cover
     def get_python_executable(self):
         raise NotImplementedError() # pragma: no cover
-    def execute_pip_install(self, *args):
+    def execute_pip_install(self, source, reinstall):
         raise NotImplementedError() # pragma: no cover
-    def execute_easy_install(self, *args):
+    def execute_easy_install(self, source, reinstall):
         raise NotImplementedError() # pragma: no cover
     def _execute(self, cmd):
         command.execute_assert_success(cmd, shell=True)
@@ -94,7 +131,17 @@ class PythonEnvironment(object):
     def _is_url(self, path_or_url):
         return bool(urlparse(path_or_url).scheme)
     def _make_source_object(self, source):
-        return Source.from_anything(source)
+        if isinstance(source, pkg_resources.Requirement):
+            specs = source.specs
+            source = source.unsafe_name
+        else:
+            specs = []
+        if isinstance(source, basestring) and source in self._aliases:
+            source = self._aliases[source]
+        if specs:
+            source = source.resolve_constraints(specs)
+        source = Source.from_anything(source)
+        return source
     def _post_install(self, source):
         raise NotImplementedError() # pragma: no cover
 
@@ -113,9 +160,9 @@ class Environment(PythonEnvironment):
         self._try_load_installed_signatures()
         virtualenv_api.activate_environment(self._path)
         set_active_environment(self)
-    def execute_pip_install(self, source):
+    def execute_pip_install(self, source, reinstall):
         self._execute("{0} {1}".format(self._get_pip_executable(), source))
-    def execute_easy_install(self, source):
+    def execute_easy_install(self, source, reinstall):
         self._execute("{0} {1}".format(self._get_easy_install_executable(), source))
     def _get_pip_executable(self):
         return os.path.join(self._get_bin_dir(), "pip")
@@ -157,10 +204,10 @@ class GlobalEnvironment(PythonEnvironment):
     _checkout_cache = None
     def create_and_activate(self):
         self._checkout_cache = CheckoutCache(self._get_pydeploy_dir())
-    def execute_easy_install(self, source):
-        return self._execute("easy_install {0}".format(source))
-    def execute_pip_install(self, source):
-        return self._execute("pip install {0}".format(source))
+    def execute_easy_install(self, source, reinstall):
+        return self._execute("easy_install {0} {1}".format("-U" if reinstall else "", source))
+    def execute_pip_install(self, source, reinstall):
+        return self._execute("pip install {0} {1}".format("-U" if reinstall else "", source))
     def get_checkout_cache(self):
         return self._checkout_cache
     def get_python_executable(self):
